/***************************************************************************
 *   Copyright (C) 2005 by Dmitry Morozhnikov   *
 *   dmiceman@mail.ru   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// for struct tm->tm_gmtoff
#define _BSD_SOURCE

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <glib.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <zlib.h>
#include <dirent.h>
#include <pthread.h>
#include <sys/statfs.h>

#include "isofs.h"

static isofs_context context;
static GHashTable *lookup_table;
static GHashTable *negative_lookup_table;
static char *negative_value = "not exists";
static pthread_mutex_t main_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t fd_mutex = PTHREAD_MUTEX_INITIALIZER;

void *isofs_real_init(char *imagefile) {
    int fd = open(imagefile, O_RDONLY);
    if(fd == -1) {
        perror("open image file");
        exit(EINVAL);
    };
    context.imagefile = imagefile;
    context.fd = fd;
    if(lseek(fd, 2048 * 16, SEEK_SET) == -1) {
        perror("can`t lseek() to primary descriptor");
        exit(EIO);
    };
    ssize_t size = read(fd, & context.pd, sizeof(context.pd));
    if(size != sizeof(context.pd)) {
        fprintf(stderr, "only %d bytes read from primary descriptor, %d required\n", size, sizeof(context.pd));
        exit(EIO);
    };
    
    context.root = (struct iso_directory_record *)& context.pd.root_directory_record;
    context.block_size = isonum_723(context.pd.logical_block_size);
    
    lookup_table = g_hash_table_new(g_str_hash, g_str_equal);
    negative_lookup_table = g_hash_table_new(g_str_hash, g_str_equal);
    
    g_hash_table_insert(lookup_table, "/", context.root);
    return &context;
};

struct iso_directory_record *isofs_lookup(const char *path) {
    if(path[0] == '\0') {
        return NULL;
    };
    struct iso_directory_record *record = g_hash_table_lookup(lookup_table, path);
    if(record) {
        return record;
    };
    if(g_hash_table_lookup(negative_lookup_table, path)) {
        return NULL;
    };
    printf("start search for %s\n", path);
    gchar **parts = g_strsplit(path, "/", -1);
    guint parts_len = g_strv_length(parts);
    int partno = 1;
    gchar *rpath = g_strdup("/");
    gchar *rpath1 = "";
    gchar *part = parts[partno];
    while(part && partno < parts_len) {
        rpath1 = g_strconcat(rpath1, "/", part, NULL);
        printf("looking for %s in %s...\n", rpath1, rpath);
        record = g_hash_table_lookup(lookup_table, rpath1);
        if(!record) {
            printf("trying to load %s...\n", rpath);
            int rc = isofs_real_readdir(rpath, NULL, NULL);
            if(rc) {
                fprintf(stderr, "lookup: error %d from readdir: %s\n", rc, strerror(-rc));
                g_strfreev(parts);
                g_free(rpath);
                return NULL;
            };
        };
        partno++;
        part = parts[partno];
        g_free(rpath);
        rpath = rpath1;
    };
    g_strfreev(parts);
    g_free(rpath);
    record = g_hash_table_lookup(lookup_table, path);
    if(!record) {
        g_hash_table_insert(negative_lookup_table, g_strdup(path), negative_value);
    };
    return record;
};

static int isofs_read_raw_block(int block, char *buf) {
    off_t off = block * context.block_size;
    if(pthread_mutex_lock(& fd_mutex)) {
        int err = errno;
        perror("readdir: can`l lock fd_mutex");
        return -err;
    };
    if(lseek(context.fd, off, SEEK_SET) == -1) {
        perror("readdir: can`t lseek()");
        pthread_mutex_unlock(& fd_mutex);
        return -EIO;
    };
    size_t len = read(context.fd, buf, context.block_size);
    if(len != context.block_size) {
        fprintf(stderr, "readdir: can`t read full block, errno %d, message %s\n", 
            errno, strerror(errno));
        pthread_mutex_unlock(& fd_mutex);
        return -EIO;
    };
    pthread_mutex_unlock(& fd_mutex);
//     printf("block %d, offset %d, read %d\n", block, (int) off, len);
    return len;
};

static int isofs_direntry2stat(struct stat *st, struct iso_directory_record *record) {
    st->st_size = isonum_733(record->size);
    st->st_blksize = context.block_size;
    st->st_blocks = st->st_size / context.block_size;
    st->st_nlink = 1;
    
    /// TODO use hidden flag?
    if(ISO_FLAGS_DIR(record->flags)) {
        st->st_mode = S_IFDIR | S_IRUSR | S_IRGRP | S_IROTH | 
            S_IXUSR | S_IXGRP | S_IXOTH; // dir, readabale and browsable by everyone
    } else {
        st->st_mode = S_IFREG | S_IRUSR | S_IRGRP | S_IROTH; // regular, read by everyone
    };
    
    struct tm tm;
    memset(& tm, 0, sizeof(struct tm));
    tm.tm_year = record->date[0];
    tm.tm_mon = record->date[1] - 1;
    tm.tm_mday = record->date[2];
    tm.tm_hour = record->date[3];
    tm.tm_min = record->date[4];
    tm.tm_sec = record->date[5];
    tm.tm_isdst = -1;
    tm.tm_gmtoff = record->date[6] * 15 * 60;
    if(tm.tm_gmtoff) {
        fprintf(stderr, "direntry2stat: non-zero timezone offset: %d\n", tm.tm_gmtoff);
    };
    time_t time = mktime(& tm);
    
    st->st_atime = time;
    st->st_ctime = time;
    st->st_mtime = time;
    
    return 0;
};

static int isofs_fix_entry(char *entry) {
    char *sep2 = index(entry, ';'); // find SEPARATOR2
    if(sep2) { // remove remaining part
        *sep2 = '\0';
    };
    char *sep1 = rindex(entry, '.'); // find SEPARATOR1
    if(sep1 && sep1[1] == '\0') { // check if SEPARATOR1 is a last symbol in string
        *sep1 = '\0'; // remove it
    };
    return 0;
};

int isofs_real_opendir(const char *path) {
    struct iso_directory_record *current = isofs_lookup(path);
    if(!current) {
        fprintf(stderr, "opendir: know nothing about %s\n", path);
        return -ENOENT;
    };
    if(!ISO_FLAGS_DIR(current->flags)) {
        fprintf(stderr, "opendir: %s not a dir\n", path);
        return -ENOTDIR;
    };
    return 0;
};

int isofs_real_readdir(const char *path, void *filler_buf, isofs_dir_fill_t filler) {
    if(path[0] == '\0') {
        fprintf(stderr, "readdir: attempt to read empty path name\n");
        return -EINVAL;
    };
    struct iso_directory_record *current = isofs_lookup(path);
    if(!current) {
        fprintf(stderr, "readdir: know nothing about %s\n", path);
        return -ENOENT;
    };
    if(!ISO_FLAGS_DIR(current->flags)) {
        fprintf(stderr, "readdir: %s not a dir\n", path);
        return -ENOTDIR;
    };
    
    size_t current_size = isonum_733(current->size);
    
    int block = isonum_733(current->extent);
    char *buf = (char *) malloc(context.block_size);
    int rc;
    
    size_t total_size = 0;
    int count = 0;
    
    while(total_size < current_size) {
        rc = isofs_read_raw_block(block, buf);
        if(rc < 0) {
            return rc;
        };
        
        off_t boff = 0;
    
        while(boff < context.block_size && total_size < current_size) {
            struct iso_directory_record *record = (struct iso_directory_record *) (buf + boff);
            count++;
            size_t record_length = isonum_711(record->length);
            size_t name_len = isonum_711(record->name_len);
    //         printf("boff %d, record length %d, name_len %d\n", (int) boff, record_length, name_len);
            
            if(record_length < sizeof(struct iso_directory_record)) {
                if(count > 2) { // check if . and .. is already read
                    // possible end of directory
                    // at least mkisofs does not set iso_directory_record.size correct
                    free(buf);
                    return 0;
                } else {
                    fprintf(stderr, "readdir: directory record length too small: %d\n", record_length);
                    free(buf);
                    return -EIO;
                };
            };
            if(name_len > 255) {
                fprintf(stderr, "readdir: file name length too big: %d\n", name_len);
                free(buf);
                return -EIO;
            };
            
            char entry[256];
            if(count == 1) { // . entry ("" on disk)
                strcpy(entry, ".");
            } else if(count == 2) { // .. entry ("" on disk)
                strcpy(entry, "..");
            } else { // regular entry
                strncpy(entry, (char *) record->name, name_len);
                entry[name_len] = '\0';
                if(!ISO_FLAGS_DIR(record->flags)) {
                    isofs_fix_entry(entry);
                };
            };
            
            if(filler) {
                struct stat st;
                memset(& st, '\0', sizeof(struct stat));
                isofs_direntry2stat(& st, record);
                rc = filler(filler_buf, entry, & st, 0);
                if(rc) {
                    free(buf);
                    return rc;
                };
            };
            
            char absolute_entry[PATH_MAX];
            strcpy(absolute_entry, path);
            if(path[1] != '\0') { // not root dir
                strcat(absolute_entry, "/");
            };
            strcat(absolute_entry, entry);
            if(g_hash_table_lookup(lookup_table, absolute_entry)) {
                // already in lookup cache
            } else {
                struct iso_directory_record *srec = 
                    (struct iso_directory_record *) malloc (sizeof(struct iso_directory_record));
                memcpy(srec, record, sizeof(struct iso_directory_record));
                g_hash_table_insert(lookup_table, g_strdup(absolute_entry), srec);
            };
            
            boff += record_length;
            total_size += record_length;
        };
    };
    
    free(buf);
    return 0;
};

int isofs_real_getattr(const char *path, struct stat *stbuf) {
    struct iso_directory_record *record = isofs_lookup(path);
    if(!record) {
        fprintf(stderr, "getattr: know nothing about %s\n", path);
        return -ENOENT;
    } else {
//         printf("getattr: found %s, size %d\n", path, isonum_733(record->size));
    };
    memset(stbuf, 0, sizeof(struct stat));
    isofs_direntry2stat(stbuf, record);
    return 0;
};

int isofs_read_block(off_t offset, size_t bsize, char *data, size_t *size) {
/*    char ibuf[PAGE_CACHE_SIZE];
    if(pthread_mutex_lock(& fd_mutex)) {
        int err = errno;
        perror("read_block: can`l lock fd_mutex");
        return -err;
    };
    if(lseek(context.fd, offset, SEEK_SET) == -1) {
        perror("read_block: can`t lseek()");
        pthread_mutex_unlock(& fd_mutex);
        return -EIO;
    };
    if(read(context.fd, ibuf, bsize) != bsize) {
        fprintf(stderr, "readdir: can`t read full block, errno %d, message %s\n", 
            errno, strerror(errno));
        pthread_mutex_unlock(& fd_mutex);
        return -EIO;
    };
    pthread_mutex_unlock(& fd_mutex);
    int rc = uncompress(data, (uLongf *) size, ibuf, bsize);
//     printf("read_block: offset %d, bsize %d, size %d, rc %d\n", (int) offset, bsize, *size, rc);
    return rc;*/
};

int isofs_real_readlink(const char *path, char *target, size_t size) {
/*    isofs_inode_context *ncontext = isofs_lookup(path);
    if(!ncontext) {
        fprintf(stderr, "readlink: know nothing about %s\n", path);
        return -ENOENT;
    };
    struct isofs_inode *inode = ncontext->inode;
    if(!S_ISLNK(inode->mode)) {
        fprintf(stderr, "readlink: %s not a link\n", path);
        return -EINVAL;
    };
    size_t fsize = inode->size;
    char *obuf = (char *) malloc(PAGE_CACHE_SIZE);
    int nblocks = (fsize - 1) / PAGE_CACHE_SIZE + 1;
    int *bbuf = (int *) malloc(nblocks * 4);
    if(pthread_mutex_lock(& fd_mutex)) {
        int err = errno;
        perror("readlink: can`l lock fd_mutex");
        return -err;
    };
    if(lseek(context.fd, inode->offset * 4, SEEK_SET) == -1) {
        perror("read_block: can`t lseek()");
        pthread_mutex_unlock(& fd_mutex);
        return -EIO;
    };
    if(read(context.fd, bbuf, nblocks * 4) != nblocks * 4) {
        fprintf(stderr, "readdir: can`t read full block table, errno %d, message %s\n", 
            errno, strerror(errno));
        pthread_mutex_unlock(& fd_mutex);
        return -EIO;
    };
    pthread_mutex_unlock(& fd_mutex);
    int i;
    off_t offset = inode->offset * 4 + nblocks * 4;
    off_t ooff = 0;
    for(i = 0; i < nblocks; i++) {
        int block = bbuf[i];
        size_t bsize = block - offset;
        if(bsize > PAGE_CACHE_SIZE) {
            free(bbuf);
            free(obuf);
            fprintf(stderr, "read: block size bigger than PAGE_CACHE_SIZE while reading block %i from symlink %s, bsize %i, offset %i, block %i\n", 
                i, path, bsize, (int) offset, block);
            return -EIO;
        };
        size_t osize = PAGE_CACHE_SIZE;
        int rc = isofs_read_block(offset, bsize, obuf + ooff, & osize);
        if(rc != Z_OK) {
            free(bbuf);
            free(obuf);
            fprintf(stderr, "readlink: read block error %i: %s\n", rc, strerror(rc));
            return -rc;
        };
        offset = block;
        ooff += osize;
        obuf[ooff] = '\0';
    };
    strncpy(target, obuf, size - 1);
    target[size - 1] = '\0';
//     printf("readlink: %s -> %s\n", path, target);
    free(bbuf);
    free(obuf);
    return 0;*/
};

int isofs_real_open(const char *path) {
    struct iso_directory_record *record = isofs_lookup(path);
    if(!record) {
        fprintf(stderr, "open: know nothing about %s\n", path);
        return -ENOENT;
    };
    if(ISO_FLAGS_DIR(record->flags)) {
        fprintf(stderr, "open: %s not a file\n", path);
        return -EINVAL;
    };
    if(ISO_FLAGS_HIDDEN(record->flags)) {
        fprintf(stderr, "open: %s is a hidden file\n", path);
        // return -EPERM;
    };
    return 0;
};

int isofs_real_read(const char *path, char *out_buf, size_t size, off_t offset) {
    struct iso_directory_record *record = isofs_lookup(path);
    if(!record) {
        fprintf(stderr, "read: know nothing about %s\n", path);
        return -ENOENT;
    };
    if(ISO_FLAGS_DIR(record->flags)) {
        fprintf(stderr, "read: %s not a file\n", path);
        return -EINVAL;
    };
    if(ISO_FLAGS_HIDDEN(record->flags)) {
        fprintf(stderr, "read: %s is a hidden file\n", path);
        // return -EPERM;
    };
    
    size_t fsize = isonum_733(record->size);
    if(offset + size > fsize) {
        size = fsize - offset;
    };
    if(size < 1) {
        return 0;
    };
    
    int start_block = isonum_733(record->extent);
    if(start_block == 0) { // empty file
        return 0;
    };
    
    int start = offset / context.block_size;
    int end = (offset + size) / context.block_size;
    int shift = offset % context.block_size;
    int block = start_block + start;
    printf("read: path %s, size %d, offset %d, fsize %d, start %d, end %d, shift %d\n", 
        path, size, (int) offset, fsize, start, end, shift);
    
    char *buf = (char *) malloc(context.block_size);
    
    int i;
    size_t total_size = 0;
    size_t size_left = size;
    int count = 0;
    
    for(i = start; i <= end; i++) {
        int len = isofs_read_raw_block(block, buf);
        if(len < 0) {
            free(buf);
            return len;
        };
        printf("read: block %d, len %d, size_left %d\n", block, len, size_left);
        
        if(len > size_left) {
            len = size_left;
        };
        
        total_size += len;
        
        memcpy(out_buf + count * context.block_size, buf + shift, len - shift);
        
        count++;
        shift = 0;
        size_left -= len;
        block++;
    };
    
    free(buf);
    return total_size;
};

int isofs_real_statfs(struct statfs *stbuf) {
    stbuf->f_type = ISOFS_SUPER_MAGIC;
    stbuf->f_bsize = context.block_size; // or PAGE_CACHE_SIZE?
    stbuf->f_blocks = 0; // while it is possible to calculate this, i see no reasons to do so
    stbuf->f_bfree = 0;
    stbuf->f_bavail = 0;
    stbuf->f_files = 0;
    stbuf->f_ffree = 0;
    stbuf->f_namelen = 255; // ? not sure..
    return 0;
};
